@@ -6,7 +6,9 @@
 #include "src/heap/factory.h"
 #include "src/heap/heap-inl.h"
 #include "src/heap/mark-compact.h"
-#include "src/heap/remembered-set.h"
+#include "src/heap/marking-state-inl.h"
+#include "src/heap/memory-chunk.h"
+#include "src/heap/remembered-set-inl.h"
 #include "src/objects/objects-inl.h"
 #include "test/cctest/cctest.h"
 #include "test/cctest/heap/heap-tester.h"
@@ -23,11 +25,7 @@ void CheckInvariantsOfAbortedPage(Page* page) {
   // 1) Markbits are cleared
   // 2) The page is not marked as evacuation candidate anymore
   // 3) The page is not marked as aborted compaction anymore.
-  CHECK(page->heap()
-            ->mark_compact_collector()
-            ->non_atomic_marking_state()
-            ->bitmap(page)
-            ->IsClean());
+  CHECK(page->heap()->non_atomic_marking_state()->bitmap(page)->IsClean());
   CHECK(!page->IsEvacuationCandidate());
   CHECK(!page->IsFlagSet(Page::COMPACTION_WAS_ABORTED));
 }
@@ -42,17 +40,23 @@ void CheckAllObjectsOnPage(const std::vector<Handle<FixedArray>>& handles,
 }  // namespace
 
 HEAP_TEST(CompactionFullAbortedPage) {
-  if (FLAG_never_compact) return;
+  if (!v8_flags.compact) return;
   // Test the scenario where we reach OOM during compaction and the whole page
   // is aborted.
 
   // Disable concurrent sweeping to ensure memory is in an expected state, i.e.,
   // we can reach the state of a half aborted page.
   ManualGCScope manual_gc_scope;
-  FLAG_manual_evacuation_candidates_selection = true;
+  heap::ManualEvacuationCandidatesSelectionScope
+      manual_evacuation_candidate_selection_scope(manual_gc_scope);
   CcTest::InitializeVM();
   Isolate* isolate = CcTest::i_isolate();
   Heap* heap = isolate->heap();
+  auto reset_oom = [](void* heap, size_t limit, size_t) -> size_t {
+    reinterpret_cast<Heap*>(heap)->set_force_oom(false);
+    return limit;
+  };
+  heap->AddNearHeapLimitCallback(reset_oom, heap);
   {
     HandleScope scope1(isolate);
 
@@ -60,7 +64,8 @@ HEAP_TEST(CompactionFullAbortedPage) {
 
     {
       HandleScope scope2(isolate);
-      CHECK(heap->old_space()->Expand());
+      CHECK(heap->old_space()->TryExpandImpl(
+          MemoryAllocator::AllocationMode::kRegular));
       auto compaction_page_handles = heap::CreatePadding(
           heap,
           static_cast<int>(MemoryChunkLayout::AllocatableMemoryInDataPage()),
@@ -72,8 +77,9 @@ HEAP_TEST(CompactionFullAbortedPage) {
       CheckAllObjectsOnPage(compaction_page_handles, to_be_aborted_page);
 
       heap->set_force_oom(true);
-      CcTest::CollectAllGarbage();
-      heap->mark_compact_collector()->EnsureSweepingCompleted();
+      heap::InvokeMajorGC(heap);
+      heap->EnsureSweepingCompleted(
+          Heap::SweepingForcedFinalizationMode::kV8Only);
 
       // Check that all handles still point to the same page, i.e., compaction
       // has been aborted on the page.
@@ -83,6 +89,7 @@ HEAP_TEST(CompactionFullAbortedPage) {
       CheckInvariantsOfAbortedPage(to_be_aborted_page);
     }
   }
+  heap->RemoveNearHeapLimitCallback(reset_oom, 0u);
 }
 
 namespace {
@@ -93,20 +100,21 @@ int GetObjectSize(int objects_per_page) {
   // Make sure that object_size is a multiple of kTaggedSize.
   int object_size =
       ((allocatable / kTaggedSize) / objects_per_page) * kTaggedSize;
-  return Min(kMaxRegularHeapObjectSize, object_size);
+  return std::min(kMaxRegularHeapObjectSize, object_size);
 }
 
 }  // namespace
 
 HEAP_TEST(CompactionPartiallyAbortedPage) {
-  if (FLAG_never_compact) return;
+  if (!v8_flags.compact) return;
   // Test the scenario where we reach OOM during compaction and parts of the
   // page have already been migrated to a new one.
 
   // Disable concurrent sweeping to ensure memory is in an expected state, i.e.,
   // we can reach the state of a half aborted page.
   ManualGCScope manual_gc_scope;
-  FLAG_manual_evacuation_candidates_selection = true;
+  heap::ManualEvacuationCandidatesSelectionScope
+      manual_evacuation_candidate_selection_scope(manual_gc_scope);
 
   const int objects_per_page = 10;
   const int object_size = GetObjectSize(objects_per_page);
@@ -114,6 +122,11 @@ HEAP_TEST(CompactionPartiallyAbortedPage) {
   CcTest::InitializeVM();
   Isolate* isolate = CcTest::i_isolate();
   Heap* heap = isolate->heap();
+  auto reset_oom = [](void* heap, size_t limit, size_t) -> size_t {
+    reinterpret_cast<Heap*>(heap)->set_force_oom(false);
+    return limit;
+  };
+  heap->AddNearHeapLimitCallback(reset_oom, heap);
   {
     HandleScope scope1(isolate);
 
@@ -123,7 +136,8 @@ HEAP_TEST(CompactionPartiallyAbortedPage) {
       HandleScope scope2(isolate);
       // Fill another page with objects of size {object_size} (last one is
       // properly adjusted).
-      CHECK(heap->old_space()->Expand());
+      CHECK(heap->old_space()->TryExpandImpl(
+          MemoryAllocator::AllocationMode::kRegular));
       auto compaction_page_handles = heap::CreatePadding(
           heap,
           static_cast<int>(MemoryChunkLayout::AllocatableMemoryInDataPage()),
@@ -138,7 +152,8 @@ HEAP_TEST(CompactionPartiallyAbortedPage) {
         // Add another page that is filled with {num_objects} objects of size
         // {object_size}.
         HandleScope scope3(isolate);
-        CHECK(heap->old_space()->Expand());
+        CHECK(heap->old_space()->TryExpandImpl(
+            MemoryAllocator::AllocationMode::kRegular));
         const int num_objects = 3;
         std::vector<Handle<FixedArray>> page_to_fill_handles =
             heap::CreatePadding(heap, object_size * num_objects,
@@ -147,8 +162,9 @@ HEAP_TEST(CompactionPartiallyAbortedPage) {
             Page::FromAddress(page_to_fill_handles.front()->address());
 
         heap->set_force_oom(true);
-        CcTest::CollectAllGarbage();
-        heap->mark_compact_collector()->EnsureSweepingCompleted();
+        heap::InvokeMajorGC(heap);
+        heap->EnsureSweepingCompleted(
+            Heap::SweepingForcedFinalizationMode::kV8Only);
 
         bool migration_aborted = false;
         for (Handle<FixedArray> object : compaction_page_handles) {
@@ -170,86 +186,11 @@ HEAP_TEST(CompactionPartiallyAbortedPage) {
       }
     }
   }
-}
-
-HEAP_TEST(CompactionPartiallyAbortedPageWithInvalidatedSlots) {
-  if (FLAG_never_compact) return;
-  // Test evacuating a page partially when it contains recorded
-  // slots and invalidated objects.
-
-  // Disable concurrent sweeping to ensure memory is in an expected state, i.e.,
-  // we can reach the state of a half aborted page.
-  ManualGCScope manual_gc_scope;
-  FLAG_manual_evacuation_candidates_selection = true;
-
-  const int objects_per_page = 10;
-  const int object_size = GetObjectSize(objects_per_page);
-
-  CcTest::InitializeVM();
-  Isolate* isolate = CcTest::i_isolate();
-  Heap* heap = isolate->heap();
-  {
-    HandleScope scope1(isolate);
-
-    heap::SealCurrentObjects(heap);
-
-    {
-      HandleScope scope2(isolate);
-      // Fill another page with objects of size {object_size} (last one is
-      // properly adjusted).
-      CHECK(heap->old_space()->Expand());
-      auto compaction_page_handles = heap::CreatePadding(
-          heap,
-          static_cast<int>(MemoryChunkLayout::AllocatableMemoryInDataPage()),
-          AllocationType::kOld, object_size);
-      Page* to_be_aborted_page =
-          Page::FromHeapObject(*compaction_page_handles.front());
-      for (Handle<FixedArray> object : compaction_page_handles) {
-        CHECK_EQ(Page::FromHeapObject(*object), to_be_aborted_page);
-
-        for (int i = 0; i < object->length(); i++) {
-          RememberedSet<OLD_TO_NEW>::Insert<AccessMode::ATOMIC>(
-              to_be_aborted_page, object->RawFieldOfElementAt(i).address());
-        }
-      }
-      // First object is going to be evacuated.
-      to_be_aborted_page->RegisterObjectWithInvalidatedSlots<OLD_TO_NEW>(
-          *compaction_page_handles.front());
-      // Last object is NOT going to be evacuated.
-      // This happens since not all objects fit on the only other page in the
-      // old space, the GC isn't allowed to allocate another page.
-      to_be_aborted_page->RegisterObjectWithInvalidatedSlots<OLD_TO_NEW>(
-          *compaction_page_handles.back());
-      to_be_aborted_page->SetFlag(
-          MemoryChunk::FORCE_EVACUATION_CANDIDATE_FOR_TESTING);
-
-      {
-        // Add another page that is filled with {num_objects} objects of size
-        // {object_size}.
-        HandleScope scope3(isolate);
-        CHECK(heap->old_space()->Expand());
-        const int num_objects = 3;
-        std::vector<Handle<FixedArray>> page_to_fill_handles =
-            heap::CreatePadding(heap, object_size * num_objects,
-                                AllocationType::kOld, object_size);
-        Page* page_to_fill =
-            Page::FromAddress(page_to_fill_handles.front()->address());
-
-        heap->set_force_oom(true);
-        CcTest::CollectAllGarbage();
-        heap->mark_compact_collector()->EnsureSweepingCompleted();
-
-        CHECK_EQ(Page::FromHeapObject(*compaction_page_handles.front()),
-                 page_to_fill);
-        CHECK_EQ(Page::FromHeapObject(*compaction_page_handles.back()),
-                 to_be_aborted_page);
-      }
-    }
-  }
+  heap->RemoveNearHeapLimitCallback(reset_oom, 0u);
 }
 
 HEAP_TEST(CompactionPartiallyAbortedPageIntraAbortedPointers) {
-  if (FLAG_never_compact) return;
+  if (!v8_flags.compact) return;
   // Test the scenario where we reach OOM during compaction and parts of the
   // page have already been migrated to a new one. Objects on the aborted page
   // are linked together. This test makes sure that intra-aborted page pointers
@@ -258,7 +199,8 @@ HEAP_TEST(CompactionPartiallyAbortedPageIntraAbortedPointers) {
   // Disable concurrent sweeping to ensure memory is in an expected state, i.e.,
   // we can reach the state of a half aborted page.
   ManualGCScope manual_gc_scope;
-  FLAG_manual_evacuation_candidates_selection = true;
+  heap::ManualEvacuationCandidatesSelectionScope
+      manual_evacuation_candidate_selection_scope(manual_gc_scope);
 
   const int objects_per_page = 10;
   const int object_size = GetObjectSize(objects_per_page);
@@ -266,6 +208,11 @@ HEAP_TEST(CompactionPartiallyAbortedPageIntraAbortedPointers) {
   CcTest::InitializeVM();
   Isolate* isolate = CcTest::i_isolate();
   Heap* heap = isolate->heap();
+  auto reset_oom = [](void* heap, size_t limit, size_t) -> size_t {
+    reinterpret_cast<Heap*>(heap)->set_force_oom(false);
+    return limit;
+  };
+  heap->AddNearHeapLimitCallback(reset_oom, heap);
   {
     HandleScope scope1(isolate);
     Handle<FixedArray> root_array =
@@ -278,7 +225,8 @@ HEAP_TEST(CompactionPartiallyAbortedPageIntraAbortedPointers) {
       HandleScope temporary_scope(isolate);
       // Fill a fresh page with objects of size {object_size} (last one is
       // properly adjusted).
-      CHECK(heap->old_space()->Expand());
+      CHECK(heap->old_space()->TryExpandImpl(
+          MemoryAllocator::AllocationMode::kRegular));
       std::vector<Handle<FixedArray>> compaction_page_handles =
           heap::CreatePadding(
               heap,
@@ -299,7 +247,8 @@ HEAP_TEST(CompactionPartiallyAbortedPageIntraAbortedPointers) {
       // Add another page that is filled with {num_objects} objects of size
       // {object_size}.
       HandleScope scope3(isolate);
-      CHECK(heap->old_space()->Expand());
+      CHECK(heap->old_space()->TryExpandImpl(
+          MemoryAllocator::AllocationMode::kRegular));
       const int num_objects = 2;
       int used_memory = object_size * num_objects;
       std::vector<Handle<FixedArray>> page_to_fill_handles =
@@ -307,9 +256,14 @@ HEAP_TEST(CompactionPartiallyAbortedPageIntraAbortedPointers) {
                               object_size);
       Page* page_to_fill = Page::FromHeapObject(*page_to_fill_handles.front());
 
+      // We need to invoke GC without stack, otherwise no compaction is
+      // performed.
+      DisableConservativeStackScanningScopeForTesting no_stack_scanning(heap);
+
       heap->set_force_oom(true);
-      CcTest::CollectAllGarbage();
-      heap->mark_compact_collector()->EnsureSweepingCompleted();
+      heap::InvokeMajorGC(heap);
+      heap->EnsureSweepingCompleted(
+          Heap::SweepingForcedFinalizationMode::kV8Only);
 
       // The following check makes sure that we compacted "some" objects, while
       // leaving others in place.
@@ -333,10 +287,11 @@ HEAP_TEST(CompactionPartiallyAbortedPageIntraAbortedPointers) {
       CheckInvariantsOfAbortedPage(to_be_aborted_page);
     }
   }
+  heap->RemoveNearHeapLimitCallback(reset_oom, 0u);
 }
 
 HEAP_TEST(CompactionPartiallyAbortedPageWithRememberedSetEntries) {
-  if (FLAG_never_compact || FLAG_always_promote_young_mc) return;
+  if (!v8_flags.compact || v8_flags.single_generation) return;
   // Test the scenario where we reach OOM during compaction and parts of the
   // page have already been migrated to a new one. Objects on the aborted page
   // are linked together and the very first object on the aborted page points
@@ -348,7 +303,8 @@ HEAP_TEST(CompactionPartiallyAbortedPageWithRememberedSetEntries) {
   // Disable concurrent sweeping to ensure memory is in an expected state, i.e.,
   // we can reach the state of a half aborted page.
   ManualGCScope manual_gc_scope;
-  FLAG_manual_evacuation_candidates_selection = true;
+  heap::ManualEvacuationCandidatesSelectionScope
+      manual_evacuation_candidate_selection_scope(manual_gc_scope);
 
   const int objects_per_page = 10;
   const int object_size = GetObjectSize(objects_per_page);
@@ -356,6 +312,11 @@ HEAP_TEST(CompactionPartiallyAbortedPageWithRememberedSetEntries) {
   CcTest::InitializeVM();
   Isolate* isolate = CcTest::i_isolate();
   Heap* heap = isolate->heap();
+  auto reset_oom = [](void* heap, size_t limit, size_t) -> size_t {
+    reinterpret_cast<Heap*>(heap)->set_force_oom(false);
+    return limit;
+  };
+  heap->AddNearHeapLimitCallback(reset_oom, heap);
   {
     HandleScope scope1(isolate);
     Handle<FixedArray> root_array =
@@ -367,7 +328,8 @@ HEAP_TEST(CompactionPartiallyAbortedPageWithRememberedSetEntries) {
       HandleScope temporary_scope(isolate);
       // Fill another page with objects of size {object_size} (last one is
       // properly adjusted).
-      CHECK(heap->old_space()->Expand());
+      CHECK(heap->old_space()->TryExpandImpl(
+          MemoryAllocator::AllocationMode::kRegular));
       auto compaction_page_handles = heap::CreatePadding(
           heap,
           static_cast<int>(MemoryChunkLayout::AllocatableMemoryInDataPage()),
@@ -394,7 +356,8 @@ HEAP_TEST(CompactionPartiallyAbortedPageWithRememberedSetEntries) {
       // Add another page that is filled with {num_objects} objects of size
       // {object_size}.
       HandleScope scope3(isolate);
-      CHECK(heap->old_space()->Expand());
+      CHECK(heap->old_space()->TryExpandImpl(
+          MemoryAllocator::AllocationMode::kRegular));
       const int num_objects = 2;
       int used_memory = object_size * num_objects;
       std::vector<Handle<FixedArray>> page_to_fill_handles =
@@ -402,9 +365,14 @@ HEAP_TEST(CompactionPartiallyAbortedPageWithRememberedSetEntries) {
                               object_size);
       Page* page_to_fill = Page::FromHeapObject(*page_to_fill_handles.front());
 
+      // We need to invoke GC without stack, otherwise no compaction is
+      // performed.
+      DisableConservativeStackScanningScopeForTesting no_stack_scanning(heap);
+
       heap->set_force_oom(true);
-      CcTest::CollectAllGarbage();
-      heap->mark_compact_collector()->EnsureSweepingCompleted();
+      heap::InvokeMajorGC(heap);
+      heap->EnsureSweepingCompleted(
+          Heap::SweepingForcedFinalizationMode::kV8Only);
 
       // The following check makes sure that we compacted "some" objects, while
       // leaving others in place.
@@ -435,7 +403,7 @@ HEAP_TEST(CompactionPartiallyAbortedPageWithRememberedSetEntries) {
       // object.
       Address broken_address = holder->address() + 2 * kTaggedSize + 1;
       // Convert it to a vector to create a string from it.
-      Vector<const uint8_t> string_to_broken_addresss(
+      base::Vector<const uint8_t> string_to_broken_addresss(
           reinterpret_cast<const uint8_t*>(&broken_address), kTaggedSize);
 
       Handle<String> string;
@@ -454,9 +422,10 @@ HEAP_TEST(CompactionPartiallyAbortedPageWithRememberedSetEntries) {
       // If remembered set entries are not properly filtered/reset for aborted
       // pages we have now a broken address at an object slot in old space and
       // the following scavenge will crash.
-      CcTest::CollectGarbage(NEW_SPACE);
+      heap::InvokeMinorGC(CcTest::heap());
     }
   }
+  heap->RemoveNearHeapLimitCallback(reset_oom, 0u);
 }
 
 }  // namespace heap
