@@ -2,6 +2,7 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
+#include "src/wasm/wasm-module-builder.h"
 #include "test/cctest/cctest.h"
 #include "test/cctest/wasm/wasm-run-utils.h"
 #include "test/common/wasm/test-signatures.h"
@@ -14,9 +15,9 @@ namespace test_run_wasm_bulk_memory {
 
 namespace {
 void CheckMemoryEquals(TestingModuleBuilder* builder, size_t index,
-                       const std::vector<byte>& expected) {
-  const byte* mem_start = builder->raw_mem_start<byte>();
-  const byte* mem_end = builder->raw_mem_end<byte>();
+                       const std::vector<uint8_t>& expected) {
+  const uint8_t* mem_start = builder->raw_mem_start<uint8_t>();
+  const uint8_t* mem_end = builder->raw_mem_end<uint8_t>();
   size_t mem_size = mem_end - mem_start;
   CHECK_LE(index, mem_size);
   CHECK_LE(index + expected.size(), mem_size);
@@ -27,8 +28,8 @@ void CheckMemoryEquals(TestingModuleBuilder* builder, size_t index,
 
 void CheckMemoryEqualsZero(TestingModuleBuilder* builder, size_t index,
                            size_t length) {
-  const byte* mem_start = builder->raw_mem_start<byte>();
-  const byte* mem_end = builder->raw_mem_end<byte>();
+  const uint8_t* mem_start = builder->raw_mem_start<uint8_t>();
+  const uint8_t* mem_end = builder->raw_mem_end<uint8_t>();
   size_t mem_size = mem_end - mem_start;
   CHECK_LE(index, mem_size);
   CHECK_LE(index + length, mem_size);
@@ -38,7 +39,7 @@ void CheckMemoryEqualsZero(TestingModuleBuilder* builder, size_t index,
 }
 
 void CheckMemoryEqualsFollowedByZeroes(TestingModuleBuilder* builder,
-                                       const std::vector<byte>& expected) {
+                                       const std::vector<uint8_t>& expected) {
   CheckMemoryEquals(builder, 0, expected);
   CheckMemoryEqualsZero(builder, expected.size(),
                         builder->mem_size() - expected.size());
@@ -46,15 +47,13 @@ void CheckMemoryEqualsFollowedByZeroes(TestingModuleBuilder* builder,
 }  // namespace
 
 WASM_EXEC_TEST(MemoryInit) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   r.builder().AddMemory(kWasmPageSize);
-  const byte data[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9};
-  r.builder().AddPassiveDataSegment(ArrayVector(data));
-  BUILD(r,
-        WASM_MEMORY_INIT(0, WASM_GET_LOCAL(0), WASM_GET_LOCAL(1),
-                         WASM_GET_LOCAL(2)),
-        kExprI32Const, 0);
+  const uint8_t data[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9};
+  r.builder().AddPassiveDataSegment(base::ArrayVector(data));
+  r.Build({WASM_MEMORY_INIT(0, WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                            WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
 
   // All zeroes.
   CheckMemoryEqualsZero(&r.builder(), 0, kWasmPageSize);
@@ -83,15 +82,13 @@ WASM_EXEC_TEST(MemoryInit) {
 }
 
 WASM_EXEC_TEST(MemoryInitOutOfBoundsData) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   r.builder().AddMemory(kWasmPageSize);
-  const byte data[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9};
-  r.builder().AddPassiveDataSegment(ArrayVector(data));
-  BUILD(r,
-        WASM_MEMORY_INIT(0, WASM_GET_LOCAL(0), WASM_GET_LOCAL(1),
-                         WASM_GET_LOCAL(2)),
-        kExprI32Const, 0);
+  const uint8_t data[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9};
+  r.builder().AddPassiveDataSegment(base::ArrayVector(data));
+  r.Build({WASM_MEMORY_INIT(0, WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                            WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
 
   const uint32_t last_5_bytes = kWasmPageSize - 5;
 
@@ -105,15 +102,13 @@ WASM_EXEC_TEST(MemoryInitOutOfBoundsData) {
 }
 
 WASM_EXEC_TEST(MemoryInitOutOfBounds) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   r.builder().AddMemory(kWasmPageSize);
-  const byte data[kWasmPageSize] = {};
-  r.builder().AddPassiveDataSegment(ArrayVector(data));
-  BUILD(r,
-        WASM_MEMORY_INIT(0, WASM_GET_LOCAL(0), WASM_GET_LOCAL(1),
-                         WASM_GET_LOCAL(2)),
-        kExprI32Const, 0);
+  const uint8_t data[kWasmPageSize] = {};
+  r.builder().AddPassiveDataSegment(base::ArrayVector(data));
+  r.Build({WASM_MEMORY_INIT(0, WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                            WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
 
   // OK, copy the full data segment to memory.
   r.Call(0, 0, kWasmPageSize);
@@ -137,15 +132,13 @@ WASM_EXEC_TEST(MemoryInitOutOfBounds) {
 }
 
 WASM_EXEC_TEST(MemoryCopy) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
-  byte* mem = r.builder().AddMemory(kWasmPageSize);
-  BUILD(
-      r,
-      WASM_MEMORY_COPY(WASM_GET_LOCAL(0), WASM_GET_LOCAL(1), WASM_GET_LOCAL(2)),
-      kExprI32Const, 0);
+  uint8_t* mem = r.builder().AddMemory(kWasmPageSize);
+  r.Build({WASM_MEMORY_COPY(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                            WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
 
-  const byte initial[] = {0, 11, 22, 33, 44, 55, 66, 77};
+  const uint8_t initial[] = {0, 11, 22, 33, 44, 55, 66, 77};
   memcpy(mem, initial, sizeof(initial));
 
   // Copy from [1, 8] to [10, 16].
@@ -166,15 +159,13 @@ WASM_EXEC_TEST(MemoryCopy) {
 }
 
 WASM_EXEC_TEST(MemoryCopyOverlapping) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
-  byte* mem = r.builder().AddMemory(kWasmPageSize);
-  BUILD(
-      r,
-      WASM_MEMORY_COPY(WASM_GET_LOCAL(0), WASM_GET_LOCAL(1), WASM_GET_LOCAL(2)),
-      kExprI32Const, 0);
+  uint8_t* mem = r.builder().AddMemory(kWasmPageSize);
+  r.Build({WASM_MEMORY_COPY(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                            WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
 
-  const byte initial[] = {10, 20, 30};
+  const uint8_t initial[] = {10, 20, 30};
   memcpy(mem, initial, sizeof(initial));
 
   // Copy from [0, 3] -> [2, 5]. The copy must not overwrite 30 before copying
@@ -189,15 +180,13 @@ WASM_EXEC_TEST(MemoryCopyOverlapping) {
 }
 
 WASM_EXEC_TEST(MemoryCopyOutOfBoundsData) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
-  byte* mem = r.builder().AddMemory(kWasmPageSize);
-  BUILD(
-      r,
-      WASM_MEMORY_COPY(WASM_GET_LOCAL(0), WASM_GET_LOCAL(1), WASM_GET_LOCAL(2)),
-      kExprI32Const, 0);
+  uint8_t* mem = r.builder().AddMemory(kWasmPageSize);
+  r.Build({WASM_MEMORY_COPY(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                            WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
 
-  const byte data[] = {11, 22, 33, 44, 55, 66, 77, 88};
+  const uint8_t data[] = {11, 22, 33, 44, 55, 66, 77, 88};
   memcpy(mem, data, sizeof(data));
 
   const uint32_t last_5_bytes = kWasmPageSize - 5;
@@ -218,13 +207,11 @@ WASM_EXEC_TEST(MemoryCopyOutOfBoundsData) {
 }
 
 WASM_EXEC_TEST(MemoryCopyOutOfBounds) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   r.builder().AddMemory(kWasmPageSize);
-  BUILD(
-      r,
-      WASM_MEMORY_COPY(WASM_GET_LOCAL(0), WASM_GET_LOCAL(1), WASM_GET_LOCAL(2)),
-      kExprI32Const, 0);
+  r.Build({WASM_MEMORY_COPY(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                            WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
 
   // Copy full range is OK.
   CHECK_EQ(0, r.Call(0, 0, kWasmPageSize));
@@ -248,13 +235,11 @@ WASM_EXEC_TEST(MemoryCopyOutOfBounds) {
 }
 
 WASM_EXEC_TEST(MemoryFill) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   r.builder().AddMemory(kWasmPageSize);
-  BUILD(
-      r,
-      WASM_MEMORY_FILL(WASM_GET_LOCAL(0), WASM_GET_LOCAL(1), WASM_GET_LOCAL(2)),
-      kExprI32Const, 0);
+  r.Build({WASM_MEMORY_FILL(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                            WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
   CHECK_EQ(0, r.Call(1, 33, 5));
   CheckMemoryEqualsFollowedByZeroes(&r.builder(), {0, 33, 33, 33, 33, 33});
 
@@ -272,42 +257,36 @@ WASM_EXEC_TEST(MemoryFill) {
 }
 
 WASM_EXEC_TEST(MemoryFillValueWrapsToByte) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   r.builder().AddMemory(kWasmPageSize);
-  BUILD(
-      r,
-      WASM_MEMORY_FILL(WASM_GET_LOCAL(0), WASM_GET_LOCAL(1), WASM_GET_LOCAL(2)),
-      kExprI32Const, 0);
+  r.Build({WASM_MEMORY_FILL(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                            WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
   CHECK_EQ(0, r.Call(0, 1000, 3));
-  const byte expected = 1000 & 255;
+  const uint8_t expected = 1000 & 255;
   CheckMemoryEqualsFollowedByZeroes(&r.builder(),
                                     {expected, expected, expected});
 }
 
 WASM_EXEC_TEST(MemoryFillOutOfBoundsData) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   r.builder().AddMemory(kWasmPageSize);
-  BUILD(
-      r,
-      WASM_MEMORY_FILL(WASM_GET_LOCAL(0), WASM_GET_LOCAL(1), WASM_GET_LOCAL(2)),
-      kExprI32Const, 0);
-  const byte v = 123;
+  r.Build({WASM_MEMORY_FILL(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                            WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
+  const uint8_t v = 123;
   CHECK_EQ(0xDEADBEEF, r.Call(kWasmPageSize - 5, v, 999));
   CheckMemoryEquals(&r.builder(), kWasmPageSize - 6, {0, 0, 0, 0, 0, 0});
 }
 
 WASM_EXEC_TEST(MemoryFillOutOfBounds) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   r.builder().AddMemory(kWasmPageSize);
-  BUILD(
-      r,
-      WASM_MEMORY_FILL(WASM_GET_LOCAL(0), WASM_GET_LOCAL(1), WASM_GET_LOCAL(2)),
-      kExprI32Const, 0);
+  r.Build({WASM_MEMORY_FILL(WASM_LOCAL_GET(0), WASM_LOCAL_GET(1),
+                            WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
 
-  const byte v = 123;
+  const uint8_t v = 123;
 
   // Destination range must not be out of bounds.
   CHECK_EQ(0xDEADBEEF, r.Call(1, v, kWasmPageSize));
@@ -322,43 +301,39 @@ WASM_EXEC_TEST(MemoryFillOutOfBounds) {
 }
 
 WASM_EXEC_TEST(DataDropTwice) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t> r(execution_tier);
   r.builder().AddMemory(kWasmPageSize);
-  const byte data[] = {0};
-  r.builder().AddPassiveDataSegment(ArrayVector(data));
-  BUILD(r, WASM_DATA_DROP(0), kExprI32Const, 0);
+  const uint8_t data[] = {0};
+  r.builder().AddPassiveDataSegment(base::ArrayVector(data));
+  r.Build({WASM_DATA_DROP(0), kExprI32Const, 0});
 
   CHECK_EQ(0, r.Call());
   CHECK_EQ(0, r.Call());
 }
 
 WASM_EXEC_TEST(DataDropThenMemoryInit) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t> r(execution_tier);
   r.builder().AddMemory(kWasmPageSize);
-  const byte data[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9};
-  r.builder().AddPassiveDataSegment(ArrayVector(data));
-  BUILD(r, WASM_DATA_DROP(0),
-        WASM_MEMORY_INIT(0, WASM_I32V_1(0), WASM_I32V_1(1), WASM_I32V_1(2)),
-        kExprI32Const, 0);
+  const uint8_t data[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9};
+  r.builder().AddPassiveDataSegment(base::ArrayVector(data));
+  r.Build({WASM_DATA_DROP(0),
+           WASM_MEMORY_INIT(0, WASM_I32V_1(0), WASM_I32V_1(1), WASM_I32V_1(2)),
+           kExprI32Const, 0});
 
   CHECK_EQ(0xDEADBEEF, r.Call());
 }
 
-void TestTableCopyInbounds(ExecutionTier execution_tier, int table_dst,
+void TestTableCopyInbounds(TestExecutionTier execution_tier, int table_dst,
                            int table_src) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   const uint32_t kTableSize = 5;
   // Add 10 function tables, even though we only test one table.
   for (int i = 0; i < 10; ++i) {
     r.builder().AddIndirectFunctionTable(nullptr, kTableSize);
   }
-  BUILD(r,
-        WASM_TABLE_COPY(table_dst, table_src, WASM_GET_LOCAL(0),
-                        WASM_GET_LOCAL(1), WASM_GET_LOCAL(2)),
-        kExprI32Const, 0);
+  r.Build({WASM_TABLE_COPY(table_dst, table_src, WASM_LOCAL_GET(0),
+                           WASM_LOCAL_GET(1), WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
 
   for (uint32_t i = 0; i <= kTableSize; ++i) {
     r.CheckCallViaJS(0, 0, 0, i);  // nop
@@ -367,22 +342,19 @@ void TestTableCopyInbounds(ExecutionTier execution_tier, int table_dst,
   }
 }
 
-WASM_EXEC_TEST(TableCopyInboundsFrom0To0) {
+WASM_COMPILED_EXEC_TEST(TableCopyInboundsFrom0To0) {
   TestTableCopyInbounds(execution_tier, 0, 0);
 }
 
-WASM_EXEC_TEST(TableCopyInboundsFrom3To0) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyInboundsFrom3To0) {
   TestTableCopyInbounds(execution_tier, 3, 0);
 }
 
-WASM_EXEC_TEST(TableCopyInboundsFrom5To9) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyInboundsFrom5To9) {
   TestTableCopyInbounds(execution_tier, 5, 9);
 }
 
-WASM_EXEC_TEST(TableCopyInboundsFrom6To6) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyInboundsFrom6To6) {
   TestTableCopyInbounds(execution_tier, 6, 6);
 }
 
@@ -410,164 +382,8 @@ void CheckTableCall(Isolate* isolate, Handle<WasmTableObject> table,
 }
 }  // namespace
 
-void TestTableInitElems(ExecutionTier execution_tier, int table_index) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
-  Isolate* isolate = CcTest::InitIsolateOnce();
-  HandleScope scope(isolate);
-  TestSignatures sigs;
-  WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
-  const uint32_t kTableSize = 5;
-  std::vector<uint32_t> function_indexes;
-  const uint32_t sig_index = r.builder().AddSignature(sigs.i_v());
-
-  for (uint32_t i = 0; i < kTableSize; ++i) {
-    WasmFunctionCompiler& fn = r.NewFunction(sigs.i_v(), "f");
-    BUILD(fn, WASM_I32V_1(i));
-    fn.SetSigIndex(sig_index);
-    function_indexes.push_back(fn.function_index());
-  }
-
-  // Passive element segment has [f0, f1, f2, f3, f4, null].
-  function_indexes.push_back(WasmElemSegment::kNullIndex);
-
-  // Add 10 function tables, even though we only test one table.
-  for (int i = 0; i < 10; ++i) {
-    r.builder().AddIndirectFunctionTable(nullptr, kTableSize);
-  }
-  r.builder().AddPassiveElementSegment(function_indexes);
-
-  WasmFunctionCompiler& call = r.NewFunction(sigs.i_i(), "call");
-  BUILD(call,
-        WASM_CALL_INDIRECT_TABLE(table_index, sig_index, WASM_GET_LOCAL(0)));
-  const uint32_t call_index = call.function_index();
-
-  BUILD(r,
-        WASM_TABLE_INIT(table_index, 0, WASM_GET_LOCAL(0), WASM_GET_LOCAL(1),
-                        WASM_GET_LOCAL(2)),
-        kExprI32Const, 0);
-
-  auto table =
-      handle(WasmTableObject::cast(
-                 r.builder().instance_object()->tables().get(table_index)),
-             isolate);
-  const double null = 0xDEADBEEF;
-
-  CheckTableCall(isolate, table, &r, call_index, null, null, null, null, null);
-
-  // 0 count is ok in bounds, and at end of regions.
-  r.CheckCallViaJS(0, 0, 0, 0);
-  r.CheckCallViaJS(0, kTableSize, 0, 0);
-  r.CheckCallViaJS(0, 0, kTableSize, 0);
-
-  // Test actual writes.
-  r.CheckCallViaJS(0, 0, 0, 1);
-  CheckTableCall(isolate, table, &r, call_index, 0, null, null, null, null);
-  r.CheckCallViaJS(0, 0, 0, 2);
-  CheckTableCall(isolate, table, &r, call_index, 0, 1, null, null, null);
-  r.CheckCallViaJS(0, 0, 0, 3);
-  CheckTableCall(isolate, table, &r, call_index, 0, 1, 2, null, null);
-  r.CheckCallViaJS(0, 3, 0, 2);
-  CheckTableCall(isolate, table, &r, call_index, 0, 1, 2, 0, 1);
-  r.CheckCallViaJS(0, 3, 1, 2);
-  CheckTableCall(isolate, table, &r, call_index, 0, 1, 2, 1, 2);
-  r.CheckCallViaJS(0, 3, 2, 2);
-  CheckTableCall(isolate, table, &r, call_index, 0, 1, 2, 2, 3);
-  r.CheckCallViaJS(0, 3, 3, 2);
-  CheckTableCall(isolate, table, &r, call_index, 0, 1, 2, 3, 4);
-}
-
-WASM_EXEC_TEST(TableInitElems0) { TestTableInitElems(execution_tier, 0); }
-WASM_EXEC_TEST(TableInitElems7) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
-  TestTableInitElems(execution_tier, 7);
-}
-WASM_EXEC_TEST(TableInitElems9) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
-  TestTableInitElems(execution_tier, 9);
-}
-
-void TestTableInitOob(ExecutionTier execution_tier, int table_index) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
-  Isolate* isolate = CcTest::InitIsolateOnce();
-  HandleScope scope(isolate);
-  TestSignatures sigs;
-  WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
-  const uint32_t kTableSize = 5;
-  std::vector<uint32_t> function_indexes;
-  const uint32_t sig_index = r.builder().AddSignature(sigs.i_v());
-
-  for (uint32_t i = 0; i < kTableSize; ++i) {
-    WasmFunctionCompiler& fn = r.NewFunction(sigs.i_v(), "f");
-    BUILD(fn, WASM_I32V_1(i));
-    fn.SetSigIndex(sig_index);
-    function_indexes.push_back(fn.function_index());
-  }
-
-  for (int i = 0; i < 10; ++i) {
-    r.builder().AddIndirectFunctionTable(nullptr, kTableSize);
-  }
-  r.builder().AddPassiveElementSegment(function_indexes);
-
-  WasmFunctionCompiler& call = r.NewFunction(sigs.i_i(), "call");
-  BUILD(call,
-        WASM_CALL_INDIRECT_TABLE(table_index, sig_index, WASM_GET_LOCAL(0)));
-  const uint32_t call_index = call.function_index();
-
-  BUILD(r,
-        WASM_TABLE_INIT(table_index, 0, WASM_GET_LOCAL(0), WASM_GET_LOCAL(1),
-                        WASM_GET_LOCAL(2)),
-        kExprI32Const, 0);
-
-  auto table =
-      handle(WasmTableObject::cast(
-                 r.builder().instance_object()->tables().get(table_index)),
-             isolate);
-  const double null = 0xDEADBEEF;
-
-  CheckTableCall(isolate, table, &r, call_index, null, null, null, null, null);
-
-  // Out-of-bounds table.init should not have any effect.
-  r.CheckCallViaJS(0xDEADBEEF, 3, 0, 3);
-  CheckTableCall(isolate, table, &r, call_index, null, null, null, null, null);
-
-  r.CheckCallViaJS(0xDEADBEEF, 0, 3, 3);
-  CheckTableCall(isolate, table, &r, call_index, null, null, null, null, null);
-
-  // 0-count is still oob if target is invalid.
-  r.CheckCallViaJS(0xDEADBEEF, kTableSize + 1, 0, 0);
-  r.CheckCallViaJS(0xDEADBEEF, 0, kTableSize + 1, 0);
-
-  r.CheckCallViaJS(0xDEADBEEF, 0, 0, 6);
-  r.CheckCallViaJS(0xDEADBEEF, 0, 1, 5);
-  r.CheckCallViaJS(0xDEADBEEF, 0, 2, 4);
-  r.CheckCallViaJS(0xDEADBEEF, 0, 3, 3);
-  r.CheckCallViaJS(0xDEADBEEF, 0, 4, 2);
-  r.CheckCallViaJS(0xDEADBEEF, 0, 5, 1);
-
-  r.CheckCallViaJS(0xDEADBEEF, 0, 0, 6);
-  r.CheckCallViaJS(0xDEADBEEF, 1, 0, 5);
-  r.CheckCallViaJS(0xDEADBEEF, 2, 0, 4);
-  r.CheckCallViaJS(0xDEADBEEF, 3, 0, 3);
-  r.CheckCallViaJS(0xDEADBEEF, 4, 0, 2);
-  r.CheckCallViaJS(0xDEADBEEF, 5, 0, 1);
-
-  r.CheckCallViaJS(0xDEADBEEF, 10, 0, 1);
-  r.CheckCallViaJS(0xDEADBEEF, 0, 10, 1);
-}
-
-WASM_EXEC_TEST(TableInitOob0) { TestTableInitOob(execution_tier, 0); }
-WASM_EXEC_TEST(TableInitOob7) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
-  TestTableInitOob(execution_tier, 7);
-}
-WASM_EXEC_TEST(TableInitOob9) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
-  TestTableInitOob(execution_tier, 9);
-}
-
-void TestTableCopyElems(ExecutionTier execution_tier, int table_dst,
+void TestTableCopyElems(TestExecutionTier execution_tier, int table_dst,
                         int table_src) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   Isolate* isolate = CcTest::InitIsolateOnce();
   HandleScope scope(isolate);
   TestSignatures sigs;
@@ -578,7 +394,7 @@ void TestTableCopyElems(ExecutionTier execution_tier, int table_dst,
 
   for (uint32_t i = 0; i < kTableSize; ++i) {
     WasmFunctionCompiler& fn = r.NewFunction(sigs.i_v(), "f");
-    BUILD(fn, WASM_I32V_1(i));
+    fn.Build({WASM_I32V_1(i)});
     fn.SetSigIndex(sig_index);
     function_indexes[i] = fn.function_index();
   }
@@ -587,12 +403,11 @@ void TestTableCopyElems(ExecutionTier execution_tier, int table_dst,
     r.builder().AddIndirectFunctionTable(function_indexes, kTableSize);
   }
 
-  BUILD(r,
-        WASM_TABLE_COPY(table_dst, table_src, WASM_GET_LOCAL(0),
-                        WASM_GET_LOCAL(1), WASM_GET_LOCAL(2)),
-        kExprI32Const, 0);
+  r.Build({WASM_TABLE_COPY(table_dst, table_src, WASM_LOCAL_GET(0),
+                           WASM_LOCAL_GET(1), WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
 
-  r.builder().FreezeSignatureMapAndInitializeWrapperCache();
+  r.builder().InitializeWrapperCache();
 
   auto table =
       handle(WasmTableObject::cast(
@@ -628,28 +443,24 @@ void TestTableCopyElems(ExecutionTier execution_tier, int table_dst,
   }
 }
 
-WASM_EXEC_TEST(TableCopyElemsFrom0To0) {
+WASM_COMPILED_EXEC_TEST(TableCopyElemsFrom0To0) {
   TestTableCopyElems(execution_tier, 0, 0);
 }
 
-WASM_EXEC_TEST(TableCopyElemsFrom3To0) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyElemsFrom3To0) {
   TestTableCopyElems(execution_tier, 3, 0);
 }
 
-WASM_EXEC_TEST(TableCopyElemsFrom5To9) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyElemsFrom5To9) {
   TestTableCopyElems(execution_tier, 5, 9);
 }
 
-WASM_EXEC_TEST(TableCopyElemsFrom6To6) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyElemsFrom6To6) {
   TestTableCopyElems(execution_tier, 6, 6);
 }
 
-void TestTableCopyCalls(ExecutionTier execution_tier, int table_dst,
+void TestTableCopyCalls(TestExecutionTier execution_tier, int table_dst,
                         int table_src) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   Isolate* isolate = CcTest::InitIsolateOnce();
   HandleScope scope(isolate);
   TestSignatures sigs;
@@ -660,7 +471,7 @@ void TestTableCopyCalls(ExecutionTier execution_tier, int table_dst,
 
   for (uint32_t i = 0; i < kTableSize; ++i) {
     WasmFunctionCompiler& fn = r.NewFunction(sigs.i_v(), "f");
-    BUILD(fn, WASM_I32V_1(i));
+    fn.Build({WASM_I32V_1(i)});
     fn.SetSigIndex(sig_index);
     function_indexes[i] = fn.function_index();
   }
@@ -670,14 +481,13 @@ void TestTableCopyCalls(ExecutionTier execution_tier, int table_dst,
   }
 
   WasmFunctionCompiler& call = r.NewFunction(sigs.i_i(), "call");
-  BUILD(call,
-        WASM_CALL_INDIRECT_TABLE(table_dst, sig_index, WASM_GET_LOCAL(0)));
+  call.Build(
+      {WASM_CALL_INDIRECT_TABLE(table_dst, sig_index, WASM_LOCAL_GET(0))});
   const uint32_t call_index = call.function_index();
 
-  BUILD(r,
-        WASM_TABLE_COPY(table_dst, table_src, WASM_GET_LOCAL(0),
-                        WASM_GET_LOCAL(1), WASM_GET_LOCAL(2)),
-        kExprI32Const, 0);
+  r.Build({WASM_TABLE_COPY(table_dst, table_src, WASM_LOCAL_GET(0),
+                           WASM_LOCAL_GET(1), WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
 
   auto table =
       handle(WasmTableObject::cast(
@@ -685,46 +495,42 @@ void TestTableCopyCalls(ExecutionTier execution_tier, int table_dst,
              isolate);
 
   if (table_dst == table_src) {
-    CheckTableCall(isolate, table, &r, call_index, 0, 1, 2, 3, 4);
+    CheckTableCall(isolate, table, &r, call_index, 0.0, 1.0, 2.0, 3.0, 4.0);
     r.CheckCallViaJS(0, 0, 1, 1);
-    CheckTableCall(isolate, table, &r, call_index, 1, 1, 2, 3, 4);
+    CheckTableCall(isolate, table, &r, call_index, 1.0, 1.0, 2.0, 3.0, 4.0);
     r.CheckCallViaJS(0, 0, 1, 2);
-    CheckTableCall(isolate, table, &r, call_index, 1, 2, 2, 3, 4);
+    CheckTableCall(isolate, table, &r, call_index, 1.0, 2.0, 2.0, 3.0, 4.0);
     r.CheckCallViaJS(0, 3, 0, 2);
-    CheckTableCall(isolate, table, &r, call_index, 1, 2, 2, 1, 2);
+    CheckTableCall(isolate, table, &r, call_index, 1.0, 2.0, 2.0, 1.0, 2.0);
   } else {
-    CheckTableCall(isolate, table, &r, call_index, 0, 1, 2, 3, 4);
+    CheckTableCall(isolate, table, &r, call_index, 0.0, 1.0, 2.0, 3.0, 4.0);
     r.CheckCallViaJS(0, 0, 1, 1);
-    CheckTableCall(isolate, table, &r, call_index, 1, 1, 2, 3, 4);
+    CheckTableCall(isolate, table, &r, call_index, 1.0, 1.0, 2.0, 3.0, 4.0);
     r.CheckCallViaJS(0, 0, 1, 2);
-    CheckTableCall(isolate, table, &r, call_index, 1, 2, 2, 3, 4);
+    CheckTableCall(isolate, table, &r, call_index, 1.0, 2.0, 2.0, 3.0, 4.0);
     r.CheckCallViaJS(0, 3, 0, 2);
-    CheckTableCall(isolate, table, &r, call_index, 1, 2, 2, 0, 1);
+    CheckTableCall(isolate, table, &r, call_index, 1.0, 2.0, 2.0, 0.0, 1.0);
   }
 }
 
-WASM_EXEC_TEST(TableCopyCallsFrom0To0) {
+WASM_COMPILED_EXEC_TEST(TableCopyCallsTo0From0) {
   TestTableCopyCalls(execution_tier, 0, 0);
 }
 
-WASM_EXEC_TEST(TableCopyCallsFrom3To0) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyCallsTo3From0) {
   TestTableCopyCalls(execution_tier, 3, 0);
 }
 
-WASM_EXEC_TEST(TableCopyCallsFrom5To9) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyCallsTo5From9) {
   TestTableCopyCalls(execution_tier, 5, 9);
 }
 
-WASM_EXEC_TEST(TableCopyCallsFrom6To6) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyCallsTo6From6) {
   TestTableCopyCalls(execution_tier, 6, 6);
 }
 
-void TestTableCopyOobWrites(ExecutionTier execution_tier, int table_dst,
+void TestTableCopyOobWrites(TestExecutionTier execution_tier, int table_dst,
                             int table_src) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   Isolate* isolate = CcTest::InitIsolateOnce();
   HandleScope scope(isolate);
   TestSignatures sigs;
@@ -735,7 +541,7 @@ void TestTableCopyOobWrites(ExecutionTier execution_tier, int table_dst,
 
   for (uint32_t i = 0; i < kTableSize; ++i) {
     WasmFunctionCompiler& fn = r.NewFunction(sigs.i_v(), "f");
-    BUILD(fn, WASM_I32V_1(i));
+    fn.Build({WASM_I32V_1(i)});
     fn.SetSigIndex(sig_index);
     function_indexes[i] = fn.function_index();
   }
@@ -744,12 +550,11 @@ void TestTableCopyOobWrites(ExecutionTier execution_tier, int table_dst,
     r.builder().AddIndirectFunctionTable(function_indexes, kTableSize);
   }
 
-  BUILD(r,
-        WASM_TABLE_COPY(table_dst, table_src, WASM_GET_LOCAL(0),
-                        WASM_GET_LOCAL(1), WASM_GET_LOCAL(2)),
-        kExprI32Const, 0);
+  r.Build({WASM_TABLE_COPY(table_dst, table_src, WASM_LOCAL_GET(0),
+                           WASM_LOCAL_GET(1), WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
 
-  r.builder().FreezeSignatureMapAndInitializeWrapperCache();
+  r.builder().InitializeWrapperCache();
 
   auto table =
       handle(WasmTableObject::cast(
@@ -779,28 +584,24 @@ void TestTableCopyOobWrites(ExecutionTier execution_tier, int table_dst,
   CheckTable(isolate, table, f0, f1, f2, f3, f4);
 }
 
-WASM_EXEC_TEST(TableCopyOobWritesFrom0To0) {
+WASM_COMPILED_EXEC_TEST(TableCopyOobWritesFrom0To0) {
   TestTableCopyOobWrites(execution_tier, 0, 0);
 }
 
-WASM_EXEC_TEST(TableCopyOobWritesFrom3To0) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyOobWritesFrom3To0) {
   TestTableCopyOobWrites(execution_tier, 3, 0);
 }
 
-WASM_EXEC_TEST(TableCopyOobWritesFrom5To9) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyOobWritesFrom5To9) {
   TestTableCopyOobWrites(execution_tier, 5, 9);
 }
 
-WASM_EXEC_TEST(TableCopyOobWritesFrom6To6) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyOobWritesFrom6To6) {
   TestTableCopyOobWrites(execution_tier, 6, 6);
 }
 
-void TestTableCopyOob1(ExecutionTier execution_tier, int table_dst,
+void TestTableCopyOob1(TestExecutionTier execution_tier, int table_dst,
                        int table_src) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
   WasmRunner<uint32_t, uint32_t, uint32_t, uint32_t> r(execution_tier);
   const uint32_t kTableSize = 5;
 
@@ -808,10 +609,9 @@ void TestTableCopyOob1(ExecutionTier execution_tier, int table_dst,
     r.builder().AddIndirectFunctionTable(nullptr, kTableSize);
   }
 
-  BUILD(r,
-        WASM_TABLE_COPY(table_dst, table_src, WASM_GET_LOCAL(0),
-                        WASM_GET_LOCAL(1), WASM_GET_LOCAL(2)),
-        kExprI32Const, 0);
+  r.Build({WASM_TABLE_COPY(table_dst, table_src, WASM_LOCAL_GET(0),
+                           WASM_LOCAL_GET(1), WASM_LOCAL_GET(2)),
+           kExprI32Const, 0});
 
   r.CheckCallViaJS(0, 0, 0, 1);           // nop
   r.CheckCallViaJS(0, 0, 0, kTableSize);  // nop
@@ -838,50 +638,22 @@ void TestTableCopyOob1(ExecutionTier execution_tier, int table_dst,
   }
 }
 
-WASM_EXEC_TEST(TableCopyOob1From0To0) {
+WASM_COMPILED_EXEC_TEST(TableCopyOob1From0To0) {
   TestTableCopyOob1(execution_tier, 0, 0);
 }
 
-WASM_EXEC_TEST(TableCopyOob1From3To0) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyOob1From3To0) {
   TestTableCopyOob1(execution_tier, 3, 0);
 }
 
-WASM_EXEC_TEST(TableCopyOob1From5To9) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyOob1From5To9) {
   TestTableCopyOob1(execution_tier, 5, 9);
 }
 
-WASM_EXEC_TEST(TableCopyOob1From6To6) {
-  EXPERIMENTAL_FLAG_SCOPE(anyref);
+WASM_COMPILED_EXEC_TEST(TableCopyOob1From6To6) {
   TestTableCopyOob1(execution_tier, 6, 6);
 }
 
-WASM_EXEC_TEST(ElemDropTwice) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
-  WasmRunner<uint32_t> r(execution_tier);
-  r.builder().AddIndirectFunctionTable(nullptr, 1);
-  r.builder().AddPassiveElementSegment({});
-  BUILD(r, WASM_ELEM_DROP(0), kExprI32Const, 0);
-
-  r.CheckCallViaJS(0);
-  r.CheckCallViaJS(0);
-}
-
-WASM_EXEC_TEST(ElemDropThenTableInit) {
-  EXPERIMENTAL_FLAG_SCOPE(bulk_memory);
-  WasmRunner<uint32_t, uint32_t> r(execution_tier);
-  r.builder().AddIndirectFunctionTable(nullptr, 1);
-  r.builder().AddPassiveElementSegment({});
-  BUILD(
-      r, WASM_ELEM_DROP(0),
-      WASM_TABLE_INIT(0, 0, WASM_I32V_1(0), WASM_I32V_1(0), WASM_GET_LOCAL(0)),
-      kExprI32Const, 0);
-
-  r.CheckCallViaJS(0, 0);
-  r.CheckCallViaJS(0xDEADBEEF, 1);
-}
-
 }  // namespace test_run_wasm_bulk_memory
 }  // namespace wasm
 }  // namespace internal
