@@ -9,25 +9,22 @@
 #include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
+
 #include <array>
 #include <memory>
 
 #include "src/base/utils/random-number-generator.h"
-#include "src/codegen/optimized-compilation-info.h"
 #include "src/compiler/compiler-source-position-table.h"
-#include "src/compiler/graph-visualizer.h"
 #include "src/compiler/int64-lowering.h"
 #include "src/compiler/js-graph.h"
 #include "src/compiler/node.h"
-#include "src/compiler/pipeline.h"
 #include "src/compiler/wasm-compiler.h"
-#include "src/compiler/zone-stats.h"
 #include "src/trap-handler/trap-handler.h"
+#include "src/wasm/canonical-types.h"
 #include "src/wasm/function-body-decoder.h"
 #include "src/wasm/local-decl-encoder.h"
 #include "src/wasm/wasm-code-manager.h"
 #include "src/wasm/wasm-external-refs.h"
-#include "src/wasm/wasm-interpreter.h"
 #include "src/wasm/wasm-js.h"
 #include "src/wasm/wasm-module.h"
 #include "src/wasm/wasm-objects-inl.h"
@@ -36,22 +33,33 @@
 #include "src/wasm/wasm-tier.h"
 #include "src/zone/accounting-allocator.h"
 #include "src/zone/zone.h"
-
 #include "test/cctest/cctest.h"
-#include "test/cctest/compiler/call-tester.h"
 #include "test/cctest/compiler/graph-and-builders.h"
-#include "test/cctest/compiler/value-helper.h"
+#include "test/common/call-tester.h"
+#include "test/common/value-helper.h"
 #include "test/common/wasm/flag-utils.h"
 
-namespace v8 {
-namespace internal {
-namespace wasm {
+namespace v8::internal::wasm {
+
+enum class TestExecutionTier : int8_t {
+  kLiftoff = static_cast<int8_t>(ExecutionTier::kLiftoff),
+  kTurbofan = static_cast<int8_t>(ExecutionTier::kTurbofan),
+  kLiftoffForFuzzing
+};
+static_assert(
+    std::is_same<std::underlying_type<ExecutionTier>::type,
+                 std::underlying_type<TestExecutionTier>::type>::value,
+    "enum types match");
+
+enum TestingModuleMemoryType { kMemory32, kMemory64 };
 
 using base::ReadLittleEndianValue;
 using base::WriteLittleEndianValue;
 
 constexpr uint32_t kMaxFunctions = 10;
 constexpr uint32_t kMaxGlobalsSize = 128;
+// Don't execute more than 16k steps.
+constexpr int kMaxNumSteps = 16 * 1024;
 
 using compiler::CallDescriptor;
 using compiler::MachineTypeForC;
@@ -62,53 +70,19 @@ using compiler::Node;
 // the trap occurs if the runtime context is not available to throw a JavaScript
 // exception.
 #define CHECK_TRAP32(x) \
-  CHECK_EQ(0xDEADBEEF, (bit_cast<uint32_t>(x)) & 0xFFFFFFFF)
-#define CHECK_TRAP64(x) \
-  CHECK_EQ(0xDEADBEEFDEADBEEF, (bit_cast<uint64_t>(x)) & 0xFFFFFFFFFFFFFFFF)
+  CHECK_EQ(0xDEADBEEF, (base::bit_cast<uint32_t>(x)) & 0xFFFFFFFF)
+#define CHECK_TRAP64(x)        \
+  CHECK_EQ(0xDEADBEEFDEADBEEF, \
+           (base::bit_cast<uint64_t>(x)) & 0xFFFFFFFFFFFFFFFF)
 #define CHECK_TRAP(x) CHECK_TRAP32(x)
 
 #define WASM_WRAPPER_RETURN_VALUE 8754
 
-#define BUILD(r, ...)                      \
-  do {                                     \
-    byte code[] = {__VA_ARGS__};           \
-    r.Build(code, code + arraysize(code)); \
-  } while (false)
-
-template <typename T>
-void AppendSingle(std::vector<byte>* code, T t) {
-  static_assert(std::is_integral<T>::value,
-                "Special types need specializations");
-  code->push_back(t);
-}
-
-// Specialized for WasmOpcode.
-template <>
-void AppendSingle<WasmOpcode>(std::vector<byte>* code, WasmOpcode op);
-
-template <typename... T>
-void Append(std::vector<byte>* code, T... ts) {
-  static_assert(sizeof...(ts) == 0, "Base case for appending bytes to code.");
-}
-
-template <typename First, typename... Rest>
-void Append(std::vector<byte>* code, First first, Rest... rest) {
-  AppendSingle(code, first);
-  Append(code, rest...);
-}
-
-// Like BUILD but pushes code bytes into a std::vector instead of an array
-// initializer. This is useful for opcodes (like SIMD), that are LEB128
-// (variable-sized). We use recursive template instantiations with variadic
-// template arguments, so that the Append calls can handle either bytes or
-// opcodes. AppendSingle is specialized for WasmOpcode, and appends multiple
-// bytes. This allows existing callers to swap out the BUILD macro for BUILD_V
-// macro without changes. Also see https://crbug.com/v8/10258.
-#define BUILD_V(r, ...)                              \
+#define ADD_CODE(vec, ...)                           \
   do {                                               \
-    std::vector<byte> code;                          \
-    Append(&code, __VA_ARGS__);                      \
-    r.Build(code.data(), code.data() + code.size()); \
+    uint8_t __buf[] = {__VA_ARGS__};                 \
+    for (size_t __i = 0; __i < sizeof(__buf); __i++) \
+      vec.push_back(__buf[__i]);                     \
   } while (false)
 
 // For tests that must manually import a JSFunction with source code.
@@ -117,25 +91,29 @@ struct ManuallyImportedJSFunction {
   Handle<JSFunction> js_function;
 };
 
+// Helper Functions.
+bool IsSameNan(float expected, float actual);
+bool IsSameNan(double expected, double actual);
+
 // A  Wasm module builder. Globals are pre-set, however, memory and code may be
 // progressively added by a test. In turn, we piecemeal update the runtime
-// objects, i.e. {WasmInstanceObject}, {WasmModuleObject} and, if necessary,
-// the interpreter.
+// objects, i.e. {WasmInstanceObject} and {WasmModuleObject}.
 class TestingModuleBuilder {
  public:
-  TestingModuleBuilder(Zone*, ManuallyImportedJSFunction*, ExecutionTier,
-                       RuntimeExceptionSupport, LowerSimd);
+  TestingModuleBuilder(Zone*, ModuleOrigin origin, ManuallyImportedJSFunction*,
+                       TestExecutionTier, RuntimeExceptionSupport,
+                       Isolate* isolate);
   ~TestingModuleBuilder();
 
-  void ChangeOriginToAsmjs() { test_module_->origin = kAsmJsSloppyOrigin; }
-
-  byte* AddMemory(uint32_t size, SharedFlag shared = SharedFlag::kNotShared);
+  uint8_t* AddMemory(uint32_t size, SharedFlag shared = SharedFlag::kNotShared,
+                     TestingModuleMemoryType = kMemory32);
 
   size_t CodeTableLength() const { return native_module_->num_functions(); }
 
   template <typename T>
-  T* AddMemoryElems(uint32_t count) {
-    AddMemory(count * sizeof(T));
+  T* AddMemoryElems(uint32_t count,
+                    TestingModuleMemoryType mem_type = kMemory32) {
+    AddMemory(count * sizeof(T), SharedFlag::kNotShared, mem_type);
     return raw_mem_start<T>();
   }
 
@@ -145,40 +123,41 @@ class TestingModuleBuilder {
     return reinterpret_cast<T*>(globals_data_ + global->offset);
   }
 
-  byte AddSignature(const FunctionSig* sig) {
-    DCHECK_EQ(test_module_->signatures.size(),
-              test_module_->signature_ids.size());
-    test_module_->signatures.push_back(sig);
-    auto canonical_sig_num = test_module_->signature_map.FindOrInsert(*sig);
-    test_module_->signature_ids.push_back(canonical_sig_num);
-    size_t size = test_module_->signatures.size();
+  // TODO(14034): Allow selecting type finality.
+  uint8_t AddSignature(const FunctionSig* sig) {
+    test_module_->add_signature(sig, kNoSuperType, v8_flags.wasm_final_types);
+    GetTypeCanonicalizer()->AddRecursiveGroup(test_module_.get(), 1);
+    instance_object_->set_isorecursive_canonical_types(
+        test_module_->isorecursive_canonical_type_ids.data());
+    size_t size = test_module_->types.size();
     CHECK_GT(127, size);
-    return static_cast<byte>(size - 1);
+    return static_cast<uint8_t>(size - 1);
   }
 
-  uint32_t mem_size() { return mem_size_; }
+  // TODO(13918): Fix the following APIs for multi-memory.
+  uint32_t mem_size() { return mem0_size_; }
 
   template <typename T>
   T* raw_mem_start() {
-    DCHECK(mem_start_);
-    return reinterpret_cast<T*>(mem_start_);
+    DCHECK_NOT_NULL(mem0_start_);
+    return reinterpret_cast<T*>(mem0_start_);
   }
 
   template <typename T>
   T* raw_mem_end() {
-    DCHECK(mem_start_);
-    return reinterpret_cast<T*>(mem_start_ + mem_size_);
+    DCHECK_NOT_NULL(mem0_start_);
+    return reinterpret_cast<T*>(mem0_start_ + mem0_size_);
   }
 
   template <typename T>
   T raw_mem_at(int i) {
-    DCHECK(mem_start_);
-    return ReadMemory(&(reinterpret_cast<T*>(mem_start_)[i]));
+    DCHECK_NOT_NULL(mem0_start_);
+    return ReadMemory(&(reinterpret_cast<T*>(mem0_start_)[i]));
   }
 
   template <typename T>
   T raw_val_at(int i) {
-    return ReadMemory(reinterpret_cast<T*>(mem_start_ + i));
+    return ReadMemory(reinterpret_cast<T*>(mem0_start_ + i));
   }
 
   template <typename T>
@@ -193,27 +172,33 @@ class TestingModuleBuilder {
 
   // Zero-initialize the memory.
   void BlankMemory() {
-    byte* raw = raw_mem_start<byte>();
-    memset(raw, 0, mem_size_);
+    uint8_t* raw = raw_mem_start<uint8_t>();
+    memset(raw, 0, mem0_size_);
   }
 
-  // Pseudo-randomly intialize the memory.
+  // Pseudo-randomly initialize the memory.
   void RandomizeMemory(unsigned int seed = 88) {
-    byte* raw = raw_mem_start<byte>();
-    byte* end = raw_mem_end<byte>();
+    uint8_t* raw = raw_mem_start<uint8_t>();
+    uint8_t* end = raw_mem_end<uint8_t>();
     v8::base::RandomNumberGenerator rng;
     rng.SetSeed(seed);
     rng.NextBytes(raw, end - raw);
   }
 
   void SetMaxMemPages(uint32_t maximum_pages) {
-    test_module_->maximum_pages = maximum_pages;
-    if (instance_object()->has_memory_object()) {
-      instance_object()->memory_object().set_maximum_pages(maximum_pages);
-    }
+    // TODO(13918): Adapt this for multi-memory.
+    DCHECK_EQ(1, test_module_->memories.size());
+    test_module_->memories[0].maximum_pages = maximum_pages;
+    DCHECK_EQ(instance_object_->memory_objects().length(),
+              test_module_->memories.size());
+    instance_object_->memory_object(0).set_maximum_pages(maximum_pages);
   }
 
-  void SetHasSharedMemory() { test_module_->has_shared_memory = true; }
+  void SetMemoryShared() {
+    // TODO(13918): Adapt this for multi-memory.
+    DCHECK_EQ(1, test_module_->memories.size());
+    test_module_->memories[0].is_shared = true;
+  }
 
   enum FunctionType { kImport, kWasm };
   uint32_t AddFunction(const FunctionSig* sig, const char* name,
@@ -221,7 +206,7 @@ class TestingModuleBuilder {
 
   // Freezes the signature map of the module and allocates the storage for
   // export wrappers.
-  void FreezeSignatureMapAndInitializeWrapperCache();
+  void InitializeWrapperCache();
 
   // Wrap the code so it can be called as a JS function.
   Handle<JSFunction> WrapCode(uint32_t index);
@@ -229,22 +214,19 @@ class TestingModuleBuilder {
   // If function_indexes is {nullptr}, the contents of the table will be
   // initialized with null functions.
   void AddIndirectFunctionTable(const uint16_t* function_indexes,
-                                uint32_t table_size);
+                                uint32_t table_size,
+                                ValueType table_type = kWasmFuncRef);
 
-  uint32_t AddBytes(Vector<const byte> bytes);
+  uint32_t AddBytes(base::Vector<const uint8_t> bytes);
 
   uint32_t AddException(const FunctionSig* sig);
 
-  uint32_t AddPassiveDataSegment(Vector<const byte> bytes);
-  uint32_t AddPassiveElementSegment(const std::vector<uint32_t>& entries);
+  uint32_t AddPassiveDataSegment(base::Vector<const uint8_t> bytes);
 
   WasmFunction* GetFunctionAt(int index) {
     return &test_module_->functions[index];
   }
 
-  WasmInterpreter* interpreter() const { return interpreter_; }
-  bool interpret() const { return interpreter_ != nullptr; }
-  LowerSimd lower_simd() const { return lower_simd_; }
   Isolate* isolate() const { return isolate_; }
   Handle<WasmInstanceObject> instance_object() const {
     return instance_object_;
@@ -256,40 +238,64 @@ class TestingModuleBuilder {
     return reinterpret_cast<Address>(globals_data_);
   }
 
-  void SetExecutable() { native_module_->SetExecutable(true); }
+  void SetDebugState() {
+    native_module_->SetDebugState(kDebugging);
+    execution_tier_ = TestExecutionTier::kLiftoff;
+  }
 
-  void TierDown() { native_module_->TierDown(isolate_); }
+  void SwitchToDebug() {
+    SetDebugState();
+    native_module_->RemoveCompiledCode(
+        NativeModule::RemoveFilter::kRemoveNonDebugCode);
+  }
 
-  enum AssumeDebugging : bool { kDebug = true, kNoDebug = false };
-  CompilationEnv CreateCompilationEnv(AssumeDebugging = kNoDebug);
+  CompilationEnv CreateCompilationEnv();
 
-  ExecutionTier execution_tier() const { return execution_tier_; }
+  TestExecutionTier test_execution_tier() const { return execution_tier_; }
+
+  ExecutionTier execution_tier() const {
+    switch (execution_tier_) {
+      case TestExecutionTier::kTurbofan:
+        return ExecutionTier::kTurbofan;
+      case TestExecutionTier::kLiftoff:
+        return ExecutionTier::kLiftoff;
+      default:
+        UNREACHABLE();
+    }
+  }
 
   RuntimeExceptionSupport runtime_exception_support() const {
     return runtime_exception_support_;
   }
 
+  void set_max_steps(int n) { max_steps_ = n; }
+  int* max_steps_ptr() { return &max_steps_; }
+  int32_t nondeterminism() { return nondeterminism_; }
+  int32_t* non_determinism_ptr() { return &nondeterminism_; }
+
+  void EnableFeature(WasmFeature feature) { enabled_features_.Add(feature); }
+
  private:
   std::shared_ptr<WasmModule> test_module_;
-  WasmModule* test_module_ptr_;
   Isolate* isolate_;
   WasmFeatures enabled_features_;
   uint32_t global_offset = 0;
-  byte* mem_start_ = nullptr;
-  uint32_t mem_size_ = 0;
-  alignas(16) byte globals_data_[kMaxGlobalsSize];
-  WasmInterpreter* interpreter_ = nullptr;
-  ExecutionTier execution_tier_;
+  // TODO(13918): Adapt for multi-memory.
+  uint8_t* mem0_start_ = nullptr;
+  uint32_t mem0_size_ = 0;
+  uint8_t* globals_data_ = nullptr;
+  TestExecutionTier execution_tier_;
   Handle<WasmInstanceObject> instance_object_;
   NativeModule* native_module_ = nullptr;
   RuntimeExceptionSupport runtime_exception_support_;
-  LowerSimd lower_simd_;
+  int32_t max_steps_ = kMaxNumSteps;
+  int32_t nondeterminism_ = 0;
 
   // Data segment arrays that are normally allocated on the instance.
-  std::vector<byte> data_segment_data_;
+  std::vector<uint8_t> data_segment_data_;
   std::vector<Address> data_segment_starts_;
   std::vector<uint32_t> data_segment_sizes_;
-  std::vector<byte> dropped_elem_segments_;
+  std::vector<uint8_t> dropped_elem_segments_;
 
   const WasmGlobal* AddGlobal(ValueType type);
 
@@ -297,23 +303,23 @@ class TestingModuleBuilder {
 };
 
 void TestBuildingGraph(Zone* zone, compiler::JSGraph* jsgraph,
-                       CompilationEnv* module, const FunctionSig* sig,
+                       CompilationEnv* env, const FunctionSig* sig,
                        compiler::SourcePositionTable* source_position_table,
-                       const byte* start, const byte* end);
+                       const uint8_t* start, const uint8_t* end);
 
 class WasmFunctionWrapper : private compiler::GraphAndBuilders {
  public:
   WasmFunctionWrapper(Zone* zone, int num_params);
 
   void Init(CallDescriptor* call_descriptor, MachineType return_type,
-            Vector<MachineType> param_types);
+            base::Vector<MachineType> param_types);
 
   template <typename ReturnType, typename... ParamTypes>
   void Init(CallDescriptor* call_descriptor) {
     std::array<MachineType, sizeof...(ParamTypes)> param_machine_types{
         {MachineTypeForC<ParamTypes>()...}};
-    Vector<MachineType> param_vec(param_machine_types.data(),
-                                  param_machine_types.size());
+    base::Vector<MachineType> param_vec(param_machine_types.data(),
+                                        param_machine_types.size());
     Init(call_descriptor, MachineTypeForC<ReturnType>(), param_vec);
   }
 
@@ -335,7 +341,7 @@ class WasmFunctionWrapper : private compiler::GraphAndBuilders {
                                        common()->HeapConstant(instance));
   }
 
-  Handle<Code> GetWrapperCode();
+  Handle<Code> GetWrapperCode(Isolate* isolate = nullptr);
 
   Signature<MachineType>* signature() const { return signature_; }
 
@@ -347,8 +353,7 @@ class WasmFunctionWrapper : private compiler::GraphAndBuilders {
 };
 
 // A helper for compiling wasm functions for testing.
-// It contains the internal state for compilation (i.e. TurboFan graph) and
-// interpretation (by adding to the interpreter manually).
+// It contains the internal state for compilation (i.e. TurboFan graph).
 class WasmFunctionCompiler : public compiler::GraphAndBuilders {
  public:
   ~WasmFunctionCompiler();
@@ -361,12 +366,16 @@ class WasmFunctionCompiler : public compiler::GraphAndBuilders {
     return descriptor_;
   }
   uint32_t function_index() { return function_->func_index; }
+  uint32_t sig_index() { return function_->sig_index; }
 
-  void Build(const byte* start, const byte* end);
+  void Build(std::initializer_list<const uint8_t> bytes) {
+    Build(base::VectorOf(bytes));
+  }
+  void Build(base::Vector<const uint8_t> bytes);
 
-  byte AllocateLocal(ValueType type) {
+  uint8_t AllocateLocal(ValueType type) {
     uint32_t index = local_decls.AddLocals(1, type);
-    byte result = static_cast<byte>(index);
+    uint8_t result = static_cast<uint8_t>(index);
     DCHECK_EQ(index, result);
     return result;
   }
@@ -387,29 +396,45 @@ class WasmFunctionCompiler : public compiler::GraphAndBuilders {
   WasmFunction* function_;
   LocalDeclEncoder local_decls;
   compiler::SourcePositionTable source_position_table_;
-  WasmInterpreter* interpreter_;
 };
 
 // A helper class to build a module around Wasm bytecode, generate machine
 // code, and run that code.
-class WasmRunnerBase : public HandleAndZoneScope {
+class WasmRunnerBase : public InitializedHandleScope {
  public:
-  WasmRunnerBase(ManuallyImportedJSFunction* maybe_import,
-                 ExecutionTier execution_tier, int num_params,
-                 RuntimeExceptionSupport runtime_exception_support,
-                 LowerSimd lower_simd)
-      : zone_(&allocator_, ZONE_NAME),
-        builder_(&zone_, maybe_import, execution_tier,
-                 runtime_exception_support, lower_simd),
+  WasmRunnerBase(ManuallyImportedJSFunction* maybe_import, ModuleOrigin origin,
+                 TestExecutionTier execution_tier, int num_params,
+                 RuntimeExceptionSupport runtime_exception_support =
+                     kNoRuntimeExceptionSupport,
+                 Isolate* isolate = nullptr)
+      : InitializedHandleScope(isolate),
+        zone_(&allocator_, ZONE_NAME, kCompressGraphZone),
+        builder_(&zone_, origin, maybe_import, execution_tier,
+                 runtime_exception_support, isolate),
         wrapper_(&zone_, num_params) {}
 
+  static void SetUpTrapCallback() {
+    WasmRunnerBase::trap_happened = false;
+    auto trap_callback = []() -> void {
+      WasmRunnerBase::trap_happened = true;
+      set_trap_callback_for_testing(nullptr);
+    };
+    set_trap_callback_for_testing(trap_callback);
+  }
+
   // Builds a graph from the given Wasm code and generates the machine
   // code and call wrapper for that graph. This method must not be called
   // more than once.
-  void Build(const byte* start, const byte* end) {
+  void Build(const uint8_t* start, const uint8_t* end) {
+    Build(base::VectorOf(start, end - start));
+  }
+  void Build(std::initializer_list<const uint8_t> bytes) {
+    Build(base::VectorOf(bytes));
+  }
+  void Build(base::Vector<const uint8_t> bytes) {
     CHECK(!compiled_);
     compiled_ = true;
-    functions_[0]->Build(start, end);
+    functions_[0]->Build(bytes);
   }
 
   // Resets the state for building the next function.
@@ -426,27 +451,22 @@ class WasmRunnerBase : public HandleAndZoneScope {
                                     const char* name = nullptr) {
     functions_.emplace_back(
         new WasmFunctionCompiler(&zone_, sig, &builder_, name));
-    builder().AddSignature(sig);
+    uint8_t sig_index = builder().AddSignature(sig);
+    functions_.back()->SetSigIndex(sig_index);
     return *functions_.back();
   }
 
-  byte AllocateLocal(ValueType type) {
+  uint8_t AllocateLocal(ValueType type) {
     return functions_[0]->AllocateLocal(type);
   }
 
   uint32_t function_index() { return functions_[0]->function_index(); }
   WasmFunction* function() { return functions_[0]->function_; }
-  WasmInterpreter* interpreter() {
-    DCHECK(interpret());
-    return functions_[0]->interpreter_;
-  }
   bool possible_nondeterminism() { return possible_nondeterminism_; }
   TestingModuleBuilder& builder() { return builder_; }
   Zone* zone() { return &zone_; }
 
-  bool interpret() { return builder_.interpret(); }
-
-  void TierDown() { builder_.TierDown(); }
+  void SwitchToDebug() { builder_.SwitchToDebug(); }
 
   template <typename ReturnType, typename... ParamTypes>
   FunctionSig* CreateSig() {
@@ -457,114 +477,11 @@ class WasmRunnerBase : public HandleAndZoneScope {
   static FunctionSig* CreateSig(Zone* zone) {
     std::array<MachineType, sizeof...(ParamTypes)> param_machine_types{
         {MachineTypeForC<ParamTypes>()...}};
-    Vector<MachineType> param_vec(param_machine_types.data(),
-                                  param_machine_types.size());
+    base::Vector<MachineType> param_vec(param_machine_types.data(),
+                                        param_machine_types.size());
     return CreateSig(zone, MachineTypeForC<ReturnType>(), param_vec);
   }
 
- private:
-  static FunctionSig* CreateSig(Zone* zone, MachineType return_type,
-                                Vector<MachineType> param_types);
-
- protected:
-  v8::internal::AccountingAllocator allocator_;
-  Zone zone_;
-  TestingModuleBuilder builder_;
-  std::vector<std::unique_ptr<WasmFunctionCompiler>> functions_;
-  WasmFunctionWrapper wrapper_;
-  bool compiled_ = false;
-  bool possible_nondeterminism_ = false;
-  int32_t main_fn_index_ = 0;
-
- public:
-  // This field has to be static. Otherwise, gcc complains about the use in
-  // the lambda context below.
-  static bool trap_happened;
-};
-
-template <typename ReturnType, typename... ParamTypes>
-class WasmRunner : public WasmRunnerBase {
- public:
-  WasmRunner(ExecutionTier execution_tier,
-             ManuallyImportedJSFunction* maybe_import = nullptr,
-             const char* main_fn_name = "main",
-             RuntimeExceptionSupport runtime_exception_support =
-                 kNoRuntimeExceptionSupport,
-             LowerSimd lower_simd = kNoLowerSimd)
-      : WasmRunnerBase(maybe_import, execution_tier, sizeof...(ParamTypes),
-                       runtime_exception_support, lower_simd) {
-    WasmFunctionCompiler& main_fn =
-        NewFunction<ReturnType, ParamTypes...>(main_fn_name);
-    // Non-zero if there is an import.
-    main_fn_index_ = main_fn.function_index();
-
-    if (!interpret()) {
-      wrapper_.Init<ReturnType, ParamTypes...>(main_fn.descriptor());
-    }
-  }
-
-  WasmRunner(ExecutionTier execution_tier, LowerSimd lower_simd)
-      : WasmRunner(execution_tier, nullptr, "main", kNoRuntimeExceptionSupport,
-                   lower_simd) {}
-
-  void SetUpTrapCallback() {
-    WasmRunnerBase::trap_happened = false;
-    auto trap_callback = []() -> void {
-      WasmRunnerBase::trap_happened = true;
-      set_trap_callback_for_testing(nullptr);
-    };
-    set_trap_callback_for_testing(trap_callback);
-  }
-
-  ReturnType Call(ParamTypes... p) {
-    DCHECK(compiled_);
-    if (interpret()) return CallInterpreter(p...);
-
-    ReturnType return_value = static_cast<ReturnType>(0xDEADBEEFDEADBEEF);
-    SetUpTrapCallback();
-
-    wrapper_.SetInnerCode(builder_.GetFunctionCode(main_fn_index_));
-    wrapper_.SetInstance(builder_.instance_object());
-    builder_.SetExecutable();
-    Handle<Code> wrapper_code = wrapper_.GetWrapperCode();
-    compiler::CodeRunner<int32_t> runner(CcTest::InitIsolateOnce(),
-                                         wrapper_code, wrapper_.signature());
-    int32_t result;
-    {
-      SetThreadInWasmFlag();
-
-      result = runner.Call(static_cast<void*>(&p)...,
-                           static_cast<void*>(&return_value));
-
-      ClearThreadInWasmFlag();
-    }
-    CHECK_EQ(WASM_WRAPPER_RETURN_VALUE, result);
-    return WasmRunnerBase::trap_happened
-               ? static_cast<ReturnType>(0xDEADBEEFDEADBEEF)
-               : return_value;
-  }
-
-  ReturnType CallInterpreter(ParamTypes... p) {
-    WasmInterpreter::Thread* thread = interpreter()->GetThread(0);
-    thread->Reset();
-    std::array<WasmValue, sizeof...(p)> args{{WasmValue(p)...}};
-    thread->InitFrame(function(), args.data());
-    thread->Run();
-    CHECK_GT(thread->NumInterpretedCalls(), 0);
-    if (thread->state() == WasmInterpreter::FINISHED) {
-      WasmValue val = thread->GetReturnValue();
-      possible_nondeterminism_ |= thread->PossibleNondeterminism();
-      return val.to<ReturnType>();
-    } else if (thread->state() == WasmInterpreter::TRAPPED) {
-      // TODO(titzer): return the correct trap code
-      int64_t result = 0xDEADBEEFDEADBEEF;
-      return static_cast<ReturnType>(result);
-    } else {
-      // TODO(titzer): falling off end
-      return ReturnType{0};
-    }
-  }
-
   void CheckCallApplyViaJS(double expected, uint32_t function_index,
                            Handle<Object>* buffer, int count) {
     Isolate* isolate = builder_.isolate();
@@ -592,10 +509,106 @@ class WasmRunner : public WasmRunnerBase {
         CHECK_DOUBLE_EQ(expected, HeapNumber::cast(*result).value());
       }
     }
+  }
 
-    if (builder_.interpret()) {
-      CHECK_GT(builder_.interpreter()->GetThread(0)->NumInterpretedCalls(), 0);
+  Handle<Code> GetWrapperCode() {
+    return wrapper_.GetWrapperCode(main_isolate());
+  }
+
+ private:
+  static FunctionSig* CreateSig(Zone* zone, MachineType return_type,
+                                base::Vector<MachineType> param_types);
+
+ protected:
+  wasm::WasmCodeRefScope code_ref_scope_;
+  std::vector<Handle<JSFunction>> jsfuncs_;
+
+  v8::internal::AccountingAllocator allocator_;
+  Zone zone_;
+  TestingModuleBuilder builder_;
+  std::vector<std::unique_ptr<WasmFunctionCompiler>> functions_;
+  WasmFunctionWrapper wrapper_;
+  bool compiled_ = false;
+  bool possible_nondeterminism_ = false;
+  int32_t main_fn_index_ = 0;
+
+  static void SetThreadInWasmFlag() {
+    *reinterpret_cast<int*>(trap_handler::GetThreadInWasmThreadLocalAddress()) =
+        true;
+  }
+
+  static void ClearThreadInWasmFlag() {
+    *reinterpret_cast<int*>(trap_handler::GetThreadInWasmThreadLocalAddress()) =
+        false;
+  }
+
+ public:
+  // This field has to be static. Otherwise, gcc complains about the use in
+  // the lambda context below.
+  static bool trap_happened;
+};
+
+template <typename T>
+inline WasmValue WasmValueInitializer(T value) {
+  return WasmValue(value);
+}
+template <>
+inline WasmValue WasmValueInitializer(int8_t value) {
+  return WasmValue(static_cast<int32_t>(value));
+}
+template <>
+inline WasmValue WasmValueInitializer(int16_t value) {
+  return WasmValue(static_cast<int32_t>(value));
+}
+
+template <typename ReturnType, typename... ParamTypes>
+class WasmRunner : public WasmRunnerBase {
+ public:
+  explicit WasmRunner(TestExecutionTier execution_tier,
+                      ModuleOrigin origin = kWasmOrigin,
+                      ManuallyImportedJSFunction* maybe_import = nullptr,
+                      const char* main_fn_name = "main",
+                      RuntimeExceptionSupport runtime_exception_support =
+                          kNoRuntimeExceptionSupport,
+                      Isolate* isolate = nullptr)
+      : WasmRunnerBase(maybe_import, origin, execution_tier,
+                       sizeof...(ParamTypes), runtime_exception_support,
+                       isolate) {
+    WasmFunctionCompiler& main_fn =
+        NewFunction<ReturnType, ParamTypes...>(main_fn_name);
+    // Non-zero if there is an import.
+    main_fn_index_ = main_fn.function_index();
+
+    wrapper_.Init<ReturnType, ParamTypes...>(main_fn.descriptor());
+  }
+
+  ReturnType Call(ParamTypes... p) {
+    DCHECK(compiled_);
+    // Save the original context, because CEntry (for runtime calls) will
+    // reset / invalidate it when returning.
+    SaveContext save_context(main_isolate());
+
+    ReturnType return_value = static_cast<ReturnType>(0xDEADBEEFDEADBEEF);
+    SetUpTrapCallback();
+
+    wrapper_.SetInnerCode(builder_.GetFunctionCode(main_fn_index_));
+    wrapper_.SetInstance(builder_.instance_object());
+    Handle<Code> wrapper_code = GetWrapperCode();
+    compiler::CodeRunner<int32_t> runner(main_isolate(), wrapper_code,
+                                         wrapper_.signature());
+    int32_t result;
+    {
+      SetThreadInWasmFlag();
+
+      result = runner.Call(static_cast<void*>(&p)...,
+                           static_cast<void*>(&return_value));
+
+      ClearThreadInWasmFlag();
     }
+    CHECK_EQ(WASM_WRAPPER_RETURN_VALUE, result);
+    return WasmRunnerBase::trap_happened
+               ? static_cast<ReturnType>(0xDEADBEEFDEADBEEF)
+               : return_value;
   }
 
   void CheckCallViaJS(double expected, ParamTypes... p) {
@@ -610,48 +623,37 @@ class WasmRunner : public WasmRunnerBase {
     CheckCallViaJS(static_cast<double>(0xDEADBEEF), p...);
   }
 
-  void CheckUsedExecutionTier(ExecutionTier expected_tier) {
-    // Liftoff can fail and fallback to Turbofan, so check that the function
-    // gets compiled by the tier requested, to guard against accidental success.
-    CHECK(compiled_);
-    CHECK_EQ(expected_tier, builder_.GetFunctionCode(0)->tier());
-  }
-
-  Handle<Code> GetWrapperCode() { return wrapper_.GetWrapperCode(); }
-
- private:
-  wasm::WasmCodeRefScope code_ref_scope_;
-  std::vector<Handle<JSFunction>> jsfuncs_;
-
-  void SetThreadInWasmFlag() {
-    *reinterpret_cast<int*>(trap_handler::GetThreadInWasmThreadLocalAddress()) =
-        true;
-  }
-
-  void ClearThreadInWasmFlag() {
-    *reinterpret_cast<int*>(trap_handler::GetThreadInWasmThreadLocalAddress()) =
-        false;
-  }
+  void SetMaxSteps(int n) { builder_.set_max_steps(n); }
+  bool HasNondeterminism() { return builder_.nondeterminism(); }
 };
 
 // A macro to define tests that run in different engine configurations.
-#define WASM_EXEC_TEST(name)                                                 \
-  void RunWasm_##name(ExecutionTier execution_tier);                         \
-  TEST(RunWasmTurbofan_##name) { RunWasm_##name(ExecutionTier::kTurbofan); } \
-  TEST(RunWasmLiftoff_##name) { RunWasm_##name(ExecutionTier::kLiftoff); }   \
-  TEST(RunWasmInterpreter_##name) {                                          \
-    RunWasm_##name(ExecutionTier::kInterpreter);                             \
-  }                                                                          \
-  void RunWasm_##name(ExecutionTier execution_tier)
+#define WASM_EXEC_TEST(name)                                                   \
+  void RunWasm_##name(TestExecutionTier execution_tier);                       \
+  TEST(RunWasmTurbofan_##name) {                                               \
+    RunWasm_##name(TestExecutionTier::kTurbofan);                              \
+  }                                                                            \
+  TEST(RunWasmLiftoff_##name) { RunWasm_##name(TestExecutionTier::kLiftoff); } \
+  void RunWasm_##name(TestExecutionTier execution_tier)
 
-#define WASM_COMPILED_EXEC_TEST(name)                                        \
-  void RunWasm_##name(ExecutionTier execution_tier);                         \
-  TEST(RunWasmTurbofan_##name) { RunWasm_##name(ExecutionTier::kTurbofan); } \
-  TEST(RunWasmLiftoff_##name) { RunWasm_##name(ExecutionTier::kLiftoff); }   \
-  void RunWasm_##name(ExecutionTier execution_tier)
+#define UNINITIALIZED_WASM_EXEC_TEST(name)               \
+  void RunWasm_##name(TestExecutionTier execution_tier); \
+  UNINITIALIZED_TEST(RunWasmTurbofan_##name) {           \
+    RunWasm_##name(TestExecutionTier::kTurbofan);        \
+  }                                                      \
+  UNINITIALIZED_TEST(RunWasmLiftoff_##name) {            \
+    RunWasm_##name(TestExecutionTier::kLiftoff);         \
+  }                                                      \
+  void RunWasm_##name(TestExecutionTier execution_tier)
 
-}  // namespace wasm
-}  // namespace internal
-}  // namespace v8
+#define WASM_COMPILED_EXEC_TEST(name)                                          \
+  void RunWasm_##name(TestExecutionTier execution_tier);                       \
+  TEST(RunWasmTurbofan_##name) {                                               \
+    RunWasm_##name(TestExecutionTier::kTurbofan);                              \
+  }                                                                            \
+  TEST(RunWasmLiftoff_##name) { RunWasm_##name(TestExecutionTier::kLiftoff); } \
+  void RunWasm_##name(TestExecutionTier execution_tier)
+
+}  // namespace v8::internal::wasm
 
 #endif
